# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from djblets.siteconfig.models import SiteConfiguration
from reviewboard.reviews.signals import (review_request_publishing,
                                         review_publishing, reply_publishing)
from reviewboard.site.urlresolvers import local_site_reverse

from rbbz.bugzilla import Bugzilla
from rbbz.errors import InvalidBugIdError, InvalidReviewerError

def publish_review_request(user, review_request_draft, **kwargs):
    try:
        bug_id = int(review_request_draft.get_bug_list()[0])
    except (IndexError, TypeError, ValueError):
        raise InvalidBugIdError

    reviewer = review_request_draft.target_people.first()
    if reviewer is None:
        raise InvalidReviewerError

    site = Site.objects.get_current()
    siteconfig = SiteConfiguration.objects.get_current()

    url = '%s://%s%s%s' % (
        siteconfig.get('site_domain_method'), site.domain,
        local_site_reverse('root').rstrip('/'),
        review_request_draft.get_review_request().get_absolute_url())

    b = Bugzilla(user.bzlogin, user.bzcookie)
    b.post_rb_url(review_request_draft.summary, bug_id, url,
                  reviewer.get_username())


review_request_publishing.connect(publish_review_request)

def publish_review(user, review, **kwargs):
    pass

review_publishing.connect(publish_review)

def publish_reply(user, reply, **kwargs):
    pass

reply_publishing.connect(publish_reply)


def get_or_create_bugzilla_users(user_data):
    # All users will have a stored password of "!", which Django uses to
    # indicate an invalid password.
    users_db = []

    for user in user_data['users']:
        username = user['email']
        real_name = user['real_name']
        can_login = user['can_login']
        modified = False

        try:
            user_db = User.objects.get(username=username)
        except User.DoesNotExist:
            user_db = User(username=username, password='!',
                           first_name=real_name, email=username,
                           is_active=can_login)
            modified = True
        else:
            if user_db.first_name != real_name:
                user_db.first_name = real_name
                modified = True

            if user_db.is_active != can_login:
                user_db.is_active = can_login
                modified = True

        if modified:
            user_db.save()

        users_db.append(user_db)
    return users_db




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from django.contrib.auth import logout
from django.contrib.auth.models import User
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.utils.translation import ugettext as _
from reviewboard.accounts.backends import AuthBackend

from rbbz.bugzilla import Bugzilla
from rbbz.errors import BugzillaAuthError, BugzillaError, BugzillaUrlError
from rbbz.forms import BugzillaAuthSettingsForm
from rbbz.models import get_or_create_bugzilla_users


class BugzillaBackend(AuthBackend):
    """
    Authenticate a user via Bugzilla XMLRPC.
    """

    backend_id = _('bugzilla')
    name = _('Bugzilla')
    login_instructions = _('using your Bugzilla credentials.')
    settings_form = BugzillaAuthSettingsForm

    def bz_error_response(self, request):
        logout(request)
        return PermissionDenied

    def authenticate(self, username, password, cookie=False):
        username = username.strip()

        try:
            bugzilla = Bugzilla()
        except BugzillaUrlError:
            return None

        user_data = bugzilla.log_in(username, password, cookie)

        if not user_data:
            return None

        users = get_or_create_bugzilla_users(user_data)

        if not users:
            return None

        user = users[0]

        if not user.is_active:
            return None

        if not cookie:
            (user.bzlogin, user.bzcookie) = bugzilla.cookies()

        return user

    def get_or_create_user(self, username, request):
        """Always check Bugzilla for updates."""
        username = username.strip()

        try:
            bugzilla = Bugzilla(*self._session_cookies(request.session))
        except BugzillaUrlError:
            return None
        except BugzillaAuthError:
            raise PermissionDenied

        user_data = bugzilla.get_user(username)

        if not user_data:
            raise self.bz_error_response(request)

        # Just store the results.
        get_or_create_bugzilla_users(user_data)

        try:
            return User.objects.get(username=username)
        except User.DoesNotExist:
            return None

    def query_users(self, query, request):
        if not query:
            return

        try:
            bugzilla = Bugzilla(*self._session_cookies(request.session))
        except BugzillaUrlError:
            return None
        except BugzillaAuthError:
            raise PermissionDenied

        try:
            get_or_create_bugzilla_users(bugzilla.query_users(query))
        except BugzillaError:
            raise PermissionDenied

    def search_users(self, query, request):
        """Search anywhere in name to support BMO :irc_nick convention."""
        q = Q(username__icontains=query)

        if request.GET.get('fullname', None):
            q = q | (Q(first_name__icontains=query) |
                     Q(last_name__icontains=query))

        return q

    def _session_cookies(self, session):
        return (session.get('Bugzilla_login'),
                session.get('Bugzilla_logincookie'))



from reviewboard.reviews.errors import PublishError

class InvalidBugIdError(PublishError):
    def __init__(self):
        PublishError.__init__(self, 'invalid bug ID')


class BugNotFoundError(PublishError):
    def __init__(self, bug_id):
        PublishError.__init__(self, 'bug %d not found' % bug_id)


class NoBugError(PublishError):
    def __init__(self):
        PublishError.__init__(self, 'no bug ID provided')


class InvalidReviewerError(PublishError):
    def __init__(self):
        PublishError.__init__(self, 'invalid reviewer')


class BugzillaError(Exception):
    def __init__(self, msg):
        self.msg = msg


class BugzillaAuthError(BugzillaError):
    pass


class BugzillaUrlError(BugzillaError):
    pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import xmlrpclib

from djblets.siteconfig.models import SiteConfiguration

from rbbz.errors import BugzillaAuthError, BugzillaError, BugzillaUrlError
from rbbz.transports import bugzilla_transport


class Bugzilla(object):
    """
    Interface to a Bugzilla system.

    At the moment this uses the XMLRPC API.  It should probably be converted
    to REST at some point.

    General FIXME: try to get more specific errors out of xmlrpclib.Fault
    exceptions.
    """

    def __init__(self, login=None, logincookie=None):
        siteconfig = SiteConfiguration.objects.get_current()
        self.xmlrpc_url = siteconfig.get('auth_bz_xmlrpc_url')

        if not self.xmlrpc_url:
            raise BugzillaUrlError

        self._transport = None
        self._proxy = None

        if logincookie:
            self.transport.set_bugzilla_cookies(login, logincookie)

    def cookies(self):
        return self.transport.bugzilla_cookies()

    def log_in(self, username, password, cookie=False):
        if cookie:
            # Username and password are actually bugzilla cookies.
            self.transport.set_bugzilla_cookies(username, password)
            user_id = username
        else:
            self.transport.remove_bugzilla_cookies()

            try:
                result = self.proxy.User.login({'login': username,
                                                'password': password})

            except xmlrpclib.Fault as e:
                if e.faultCode == 300 or e.faultCode == 301:
                    # Invalid username/password or account disabled
                    return None
                raise BugzillaError(e.faultString)

            user_id = result['id']

        params = {'ids': [user_id],
                  'include_fields': ['email', 'real_name', 'can_login']}

        try:
            return self.proxy.User.get(params)
        except xmlrpclib.Fault as e:
            raise BugzillaError(e.faultString)

    def get_user(self, username):
        params = {'names': [username],
                  'include_fields': ['email', 'real_name', 'can_login']}

        try:
            return self.proxy.User.get(params)
        except xmlrpclib.Fault:
            raise BugzillaError

    def query_users(self, query):
        params = {'match': [query],
                  'include_fields': ['email', 'real_name', 'can_login']}

        try:
            return self.proxy.User.get(params)
        except xmlrpclib.Fault:
            raise BugzillaError

    def post_comment(self, bug_id, comment):
        params = {
            'id': bug_id,
            'comment': comment
        }

        try:
            return self.proxy.Bug.add_comment(params)
        except xmlrpclib.Fault:
            raise BugzillaError

    def post_rb_url(self, summary, bug_id, url, reviewer):
        params = {
            'ids': [bug_id],
            'data': url,
            'file_name': summary,
            'summary': summary,
            'content_type': 'text/plain',
            'flags': [{'name': 'review',
                       'status': '?',
                       'requestee': reviewer}]
        }

        try:
            return self.proxy.Bug.add_attachment(params)
        except xmlrpclib.Fault:
            raise BugzillaError

    @property
    def transport(self):
        if self._transport is None:
            self._transport = bugzilla_transport(self.xmlrpc_url)

        return self._transport

    @property
    def proxy(self):
        if self._proxy is None:
            self._proxy = xmlrpclib.ServerProxy(self.xmlrpc_url, self.transport)

        return self._proxy



from setuptools import setup

PACKAGE = 'rbbz'
VERSION = '0.1.1'

setup(name=PACKAGE,
      version=VERSION,
      description='Review Board extension for Bugzilla support',
      url='https://github.com/mozilla/rbbz',
      author='Mark Cote',
      author_email='mcote@mozilla.com',
      license='MPL 2.0',
      packages=['rbbz'],
      entry_points={
        'reviewboard.extensions':
            '%s = rbbz.extension:BugzillaExtension' % PACKAGE,
        'reviewboard.auth_backends': [
            'bugzilla = rbbz.auth:BugzillaBackend',
        ],
      }
)
